#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <wchar.h>
#include <algorithm>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Detector*         detector;
extern Options           op;

Manager::Manager(const wchar_t* fileName, const FmtPlugin *fmtPlugin)
{
  lstrcpyW(hostFileName, fileName);
  fmt = fmtPlugin;
  noFiles      = 0;
  noDelFiles   = 0;
  curFolderNum = 0;
  *curFolder   = 0;
  
  img = fmt->init(hostFileName);
  ZeroMemory(&lastModifed, sizeof(lastModifed));

  ZeroMemory(fileMap,   sizeof(fileMap));
  ZeroMemory(folderMap, sizeof(folderMap));
}

Manager::~Manager()
{
  fmt->cleanup(img);
}

void Manager::GetOpenPanelInfo(struct OpenPanelInfo *Info)
{
  Info->StructSize    = sizeof(*Info);
  Info->Flags         = OPIF_SHOWNAMESONLY | OPIF_SHOWPRESERVECASE;
  Info->HostFile      = hostFileName;
  Info->CurDir        = curFolder;
  Info->Format        = L"TR-DOS disk";
  Info->StartSortMode = SM_UNSORTED;
  
  static wchar_t panelTitle[260];
  panelTitle[0] = 0;
  lstrcatW(panelTitle, L" xTRD:");
  lstrcatW(panelTitle, pointToName(hostFileName));
  if(op.useDS) lstrcatW(panelTitle, curFolder);
  lstrcatW(panelTitle, L" ");
  Info->PanelTitle = panelTitle;

  static PanelMode mode[10];
  static const wchar_t*     columnTitles3[4];
  columnTitles3[0] = GetMsg(MName);
  columnTitles3[1] = GetMsg(MSectorSize);
  columnTitles3[2] = GetMsg(MName);
  columnTitles3[3] = GetMsg(MSectorSize);

  static const wchar_t*     columnTitles4[6];
  columnTitles4[0] = GetMsg(MName);
  columnTitles4[1] = GetMsg(MSize);
  columnTitles4[2] = GetMsg(MStart);
  columnTitles4[3] = GetMsg(MSectorSize);
  columnTitles4[4] = GetMsg(MTrack);
  columnTitles4[5] = GetMsg(MSector);

  static const wchar_t*     columnTitles5[2];
  columnTitles5[0] = GetMsg(MName);
  columnTitles5[1] = GetMsg(MFormat);

  static const wchar_t*     columnTitles6[2];
  columnTitles6[0] = GetMsg(MName);
  columnTitles6[1] = GetMsg(MComment);
  
  // C0 - trdos name
  // C1 - trdos 'start'
  // C2 -   
  // 3 - 
  // C4 - 
  // C5 -  
  // C6 -   
  
  mode[3].ColumnTypes=L"N,C2,N,C2";
  mode[3].ColumnWidths=L"0,3,0,3";
  mode[3].ColumnTitles=columnTitles3;
  mode[3].StatusColumnTypes=L"C0,S,C2";
  mode[3].StatusColumnWidths=L"0,6,3";
  mode[3].Flags = PMFLAGS_ALIGNEXTENSIONS;
  
  mode[4].ColumnTypes=L"C0,S,C1,C2,C3,C4";
  mode[4].ColumnWidths=L"0,6,5,3,3,3";
  mode[4].StatusColumnTypes=L"N,S,C2";
  mode[4].StatusColumnWidths=L"0,6,3";
  mode[4].ColumnTitles=columnTitles4;
  
  mode[5].ColumnTypes=L"C0,C5";
  mode[5].ColumnWidths=L"12,0";
  mode[5].StatusColumnTypes=L"N,S,C2";
  mode[5].StatusColumnWidths=L"0,6,3";
  mode[5].ColumnTitles=columnTitles5;

  mode[6].ColumnTypes=L"C0,C6";
  mode[6].ColumnWidths=L"12,0";
  mode[6].StatusColumnTypes=L"N,S,C2";
  mode[6].StatusColumnWidths=L"0,6,3";
  mode[6].ColumnTitles=columnTitles6;
  
  Info->PanelModesArray = mode;
  Info->PanelModesNumber = std::size(mode);
  Info->StartPanelMode = '0' + op.defaultPanelMode;
  if(readInfo())
  {
    static InfoPanelLine infoLines[14];
    infoLines[0].Text = GetMsg(MDiskInfo);
    infoLines[0].Flags = IPLFLAGS_SEPARATOR;

    infoLines[1].Text = GetMsg(MImageType);
    infoLines[1].Data = fmt->getFormatName();
    infoLines[1].Flags = 0;

    infoLines[2].Text = GetMsg(MDiskTitle);
    int noChars = 11;
    while(noChars--)
      if(diskInfo.title[noChars] != ' ' && diskInfo.title[noChars] != 0) break;

    static wchar_t Title2[260];
    swprintf(Title2, L"%.*hs", noChars+2, diskInfo.title);
    infoLines[2].Data = Title2;
    infoLines[2].Flags = 0;

    infoLines[3].Text = GetMsg(MDiskType);
    switch(diskInfo.type)
    {
      case 0x16:
                infoLines[3].Data = GetMsg(M2SDD);
                break;
      case 0x17:
                infoLines[3].Data = GetMsg(M2SSD);
                break;
      case 0x18:
                infoLines[3].Data = GetMsg(M1SDD);
                break;
      case 0x19:
                infoLines[3].Data = GetMsg(M1SSD);
                break;
      default:
                infoLines[3].Data = GetMsg(MDiskUnknown);
                break;
    }
    infoLines[3].Flags = 0;

    infoLines[4].Text = GetMsg(MWriteProtection);
    if(fmt->isProtected(img))
      infoLines[4].Data = GetMsg(MOn);
    else
      infoLines[4].Data = GetMsg(MOff);
    infoLines[4].Flags = 0;
    
    infoLines[5].Text = GetMsg(MFilesInfo);
    infoLines[5].Flags = IPLFLAGS_SEPARATOR;
    
    infoLines[6].Text = GetMsg(MFiles);
    static wchar_t noFilesStr[20];
    wsprintfW(noFilesStr, L"%d", diskInfo.noFiles);
    infoLines[6].Data = noFilesStr;
    infoLines[6].Flags = 0;
    infoLines[7].Text = GetMsg(MFilesDel);
    static wchar_t noDelFilesStr[20];
    wsprintfW(noDelFilesStr, L"%d", diskInfo.noDelFiles);
    infoLines[7].Data = noDelFilesStr;
    infoLines[7].Flags = 0;
    
    infoLines[8].Text = GetMsg(MDirInfo);
    infoLines[8].Flags = IPLFLAGS_SEPARATOR;
    
    infoLines[9].Text = GetMsg(MDirSys);
    if(!op.useDS)
      infoLines[9].Data = GetMsg(MDisabled);
    else
    {
      if(!dsOk)
        infoLines[9].Data = GetMsg(MAbsent);
      else
        infoLines[9].Data = GetMsg(MDS100);
    }
    infoLines[9].Flags = 0;
    
    infoLines[10].Text = GetMsg(MFreeAreaInfo);
    infoLines[10].Flags = IPLFLAGS_SEPARATOR;
    
    infoLines[11].Text = GetMsg(M1Trk);
    static wchar_t firstFreeTrk[20];
    wsprintfW(firstFreeTrk, L"%d", diskInfo.firstFreeTrk);
    infoLines[11].Data = firstFreeTrk;
    infoLines[11].Flags = 0;

    infoLines[12].Text = GetMsg(M1Sec);
    static wchar_t firstFreeSec[20];
    wsprintfW(firstFreeSec, L"%d", diskInfo.firstFreeSec);
    infoLines[12].Data = firstFreeSec;
    infoLines[12].Flags = 0;

    infoLines[13].Text = GetMsg(MFreeSecs);
    static wchar_t noFreeSecs[20];
    wsprintfW(noFreeSecs, L"%d", diskInfo.noFreeSecs);
    infoLines[13].Data  = noFreeSecs;
    infoLines[13].Flags = 0;

    Info->InfoLines = infoLines;
    Info->InfoLinesNumber = std::size(infoLines);
  }
}

int Manager::getFindData(PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode)
{
  if(!readInfo()) return FALSE;
  //        
  int noItems;
  int noFiles   = 0;
  int noFolders = 0;
  if(!op.useDS || !dsOk)
  {
    noItems = noFiles = this->noFiles;
  }
  else
  {
    for(int i = 0; i < this->noFolders; ++i)
      if(folderMap[i] == curFolderNum) ++noFolders;
    for(int i = 0; i < this->noFiles; ++i)
      if(fileMap[i] == curFolderNum) ++noFiles;
    noItems = noFolders + noFiles;
  }
  
  PluginPanelItem *item = new PluginPanelItem[noItems + 1]{};

  *pNoItems   = noItems+1;
  *pPanelItem = item;

  item[0].FileName = L"..";
  item[0].FileAttributes = FILE_ATTRIBUTE_DIRECTORY;

  item[0].CustomColumnNumber  = 1;
  wchar_t **CustomColumnData = new LPWSTR[item[0].CustomColumnNumber];
  item[0].CustomColumnData    = CustomColumnData;
  CustomColumnData[0] = new wchar_t[3];
  lstrcpyW(CustomColumnData[0], L"..");

  if(noItems == 0) return TRUE;

  //      trdos' 
  int nameColumnWidth = 12;
  if((opMode & OPM_FIND) == 0)
  {
    //     ,    
    //  if'   
    PanelInfo panel = { sizeof(PanelInfo) };
    if(!startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &panel))
    {
        return FALSE;
    }

    if(panel.ViewMode == 4)
      // 24 -   ,  2 -  
      nameColumnWidth = panel.PanelRect.right - panel.PanelRect.left - 2 - 24;
    if(nameColumnWidth < 12) nameColumnWidth = 12;
  }
  
  wchar_t* name = new wchar_t[nameColumnWidth+1];
  if(op.useDS && dsOk)
  {
    int folderNum = 0;
    for(int i = 0; i < noFolders; ++i)
    {
      int j = folderNum;
      for(; j < sizeof(folderMap); ++j) if(folderMap[j] == curFolderNum) break;

      folderNum = j;
      item[i+1].FileAttributes = folders[folderNum][0] != 0x01 ? FILE_ATTRIBUTE_DIRECTORY : FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN;
      item[i+1].FileName = pcFolders[folderNum];

      item[i+1].CustomColumnNumber  = 1;
      wchar_t **CustomColumnData    = new LPWSTR[item[i+1].CustomColumnNumber];
      item[i+1].CustomColumnData    = CustomColumnData;
      CustomColumnData[0] = new wchar_t[nameColumnWidth+1];
      wmemset(name, L' ', nameColumnWidth+1);

      std::transform(folders[folderNum], folders[folderNum] + 8, name, Cp866ToUcs2);
      std::transform(folders[folderNum] + 8, folders[folderNum] + 8 + 3, name+nameColumnWidth-3, Cp866ToUcs2);
      name[nameColumnWidth] = 0;
      lstrcpyW(CustomColumnData[0], name);
      ++folderNum;
    }
  }
  int fNum = 0;
  int iNum;
  for(int k = 0; k < noFiles; ++k)
  {
    if(!op.useDS || !dsOk)
    {
      iNum = k + 1;
      fNum = k;
    }
    else
    {
      iNum = noFolders + k + 1;
      int j = fNum;
      for(; j < sizeof(fileMap); ++j) if(fileMap[j] == curFolderNum) break;
      fNum = j;
    }
    
    item[iNum].FileAttributes = files[fNum].name[0] == 0x01 ? FILE_ATTRIBUTE_HIDDEN : FILE_ATTRIBUTE_NORMAL;
    item[iNum].FileName = pcFiles[fNum].name;
    item[iNum].FileSize = files[fNum].size;
    
    item[iNum].CustomColumnNumber = 7;
    wchar_t **CustomColumnData7 = new LPWSTR[item[iNum].CustomColumnNumber];
    item[iNum].CustomColumnData = CustomColumnData7;
    
    CustomColumnData7[0] = new wchar_t[nameColumnWidth+1];

    wmemset(name, L' ', nameColumnWidth+1);

    makeTrDosName(name, files[fNum], op.showExt, nameColumnWidth);
    lstrcpyW(CustomColumnData7[0], name);
    
    CustomColumnData7[1] = new wchar_t[6];
    wsprintfW(CustomColumnData7[1], L"%5d", files[fNum].start);

    CustomColumnData7[2] = new wchar_t[4];
    wsprintfW(CustomColumnData7[2],
             L"%3d",
             files[fNum].noSecs);

    CustomColumnData7[3] = new wchar_t[4];
    wsprintfW(CustomColumnData7[3], L"%3d", files[fNum].trk);

    CustomColumnData7[4] = new wchar_t[4];
    wsprintfW(CustomColumnData7[4], L"%3d", files[fNum].sec);

    
    const wchar_t* description = detector->description(pcFiles[fNum].type);
    if(description)
    {
      CustomColumnData7[5] = new wchar_t[lstrlenW(description)+1];
      lstrcpyW(CustomColumnData7[5], description);
    }
    else
    {
      CustomColumnData7[5] = new wchar_t[1];
      CustomColumnData7[5][0] = 0;
    }
    const wchar_t* comment = pcFiles[fNum].comment;
    if(*comment)
    {
      CustomColumnData7[6] = new wchar_t[lstrlenW(comment)+1];
      lstrcpyW(CustomColumnData7[6], comment);
    }
    else
    {
      CustomColumnData7[6] = new wchar_t[1];
      CustomColumnData7[6][0] = 0;
    }
    ++fNum;
  }
  delete[] name;
  return TRUE;
}

void Manager::freeFindData(PluginPanelItem *panelItem, int  noItems)
{
  for (int i = 0; i < noItems; i++)
  {
    for(int j = 0; j < panelItem[i].CustomColumnNumber; ++j)
      delete[] panelItem[i].CustomColumnData[j];
    delete[] panelItem[i].CustomColumnData;
  }
  delete[] panelItem;
}

int Manager::processHostFile(PluginPanelItem* panelItem, int noItems, int opMode)
{
  return diskMenu();
}
