#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdlib.h>
#include <wchar.h>
#include <memory>
#include <string>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

void addEndSlash(std::wstring &path);

namespace
{
  enum Action { ASK_USER, WRITE_ALL, SKIP_ALL };
  Action userAction;

  bool   keepSilence;
}

bool Manager::isDiskFull(int noAddFiles, int noSecs)
{
  if(noFiles + noAddFiles > 128)
  {
    const wchar_t * const msgItems[]{ GetMsg(MDiskFull), GetMsg(MTooManyFiles), GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return true;
  }
  if(noSecs > diskInfo.noFreeSecs)
  {
    wchar_t str1[80];
    wchar_t str2[80];
    wsprintfW(str1, GetMsg(MFileSize), noSecs);
    wsprintfW(str2, GetMsg(MFree), diskInfo.noFreeSecs);
    const wchar_t * const msgItems[]{ GetMsg(MDiskFull), str1, str2, GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return true;
  }
  return false;
}

ExitCode Manager::writeFile(const SCLHdr& h, HANDLE file, BYTE folderNum)
{
  if(isDiskFull(1, h.noSecs)) return CANCEL;
  
  FileHdr hdr;
  memcpy(&hdr, &h, sizeof(SCLHdr));
  hdr.trk = diskInfo.firstFreeTrk;
  hdr.sec = diskInfo.firstFreeSec;
  
  int fNum;
  for(fNum = 0; fNum < noFiles; ++fNum)
  {
    if(dsOk && fileMap[fNum] != folderNum) continue;
    static_assert(_countof(files[fNum].name) == std::size(hdr.name));
    if(std::equal(std::cbegin(files[fNum].name), std::cend(files[fNum].name), hdr.name)) break;
  }
  if(fNum != noFiles)
  {
    if(userAction == SKIP_ALL) return SKIP;
    if(userAction == ASK_USER)
    {
      const wchar_t *msgItems[13];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[3] = L"\x001";
      msgItems[4] = GetMsg(MLine);
      msgItems[7] = L"\x001";
      msgItems[8] = GetMsg(MWrite);
      msgItems[9] = GetMsg(MAll);
      msgItems[10] = GetMsg(MSkip);
      msgItems[11] = GetMsg(MSkipAll);
      msgItems[12] = GetMsg(MCancel);

      wchar_t name[13] = L"            ";
      makeTrDosName(name, hdr, op.showExt, 12);
      msgItems[2] = name;

      wchar_t source[50];
      wchar_t dest[50];
      wsprintfW(source, GetMsg(MSource),
               (int)hdr.size,
               hdr.noSecs);
      wsprintfW(dest,   GetMsg(MDestination),
               (int)files[fNum].size,
               files[fNum].noSecs);
      
      msgItems[5] = source;
      msgItems[6] = dest;
      
      int askCode = messageBox(FMSG_WARNING/* | FMSG_DOWN*/,
                               msgItems,
                               std::size(msgItems), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return CANCEL;
        case 3:
                  userAction = SKIP_ALL;
        case 2:
                  {
                    //     SCL  PLAIN 
                    SetFilePointer(file, hdr.noSecs*sectorSize, NULL, FILE_CURRENT);
                    return SKIP;
                  }
        case 1:
                  userAction = WRITE_ALL;
      }
    }
  }
  //   
  int trk = diskInfo.firstFreeTrk;
  int sec = diskInfo.firstFreeSec;

  if(hdr.noSecs)
  {
    WORD noSecs = hdr.noSecs;
    while(noSecs--)
    {
      BYTE sector[sectorSize];
      ZeroMemory(&sector, sizeof(sector));
      DWORD noBytesRead;
      ReadFile(file, sector, sectorSize, &noBytesRead, NULL);
      write(trk, sec, sector);
      if(++sec == 16) { ++trk; sec = 0; }
    }
  }
  diskInfo.firstFreeTrk = trk;
  diskInfo.firstFreeSec = sec;

  if(dsOk) fileMap[noFiles] = folderNum;
  
  files[noFiles++] = hdr;
  if(hdr.name[0] == 0x01) ++noDelFiles;
  diskInfo.noFreeSecs -= hdr.noSecs;
  
  return OK;
}

ExitCode Manager::putFile(const wchar_t* fileName, BYTE folderNum, bool move)
{
  wchar_t shortName[50];
  const wchar_t * const msgItems[]{ GetMsg(MCopyingFile), GetMsg(MCopyTheFile), truncPathStr(shortName, fileName, 30) };

  if(!keepSilence)
    messageBox(0, msgItems, std::size(msgItems), 0);

  HANDLE file = CreateFileW(fileName,
                           GENERIC_READ,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_SEQUENTIAL_SCAN,
                           NULL);

  if(file == INVALID_HANDLE_VALUE)
  {
    const wchar_t * const msgItems[]{ GetMsg(MError), GetMsg(MCanNotOpen), fileName, GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return CANCEL;
  }

  DWORD noBytesRead;
  
  ExitCode exitCode = OK;
  FileType type     = detectFileType(file);
  
  if(type == FMT_HOBETA)
  {
    HoHdr hoHdr;
    ReadFile(file, &hoHdr, sizeof(HoHdr), &noBytesRead, 0);

    SCLHdr h;
    memcpy(&h, &hoHdr, sizeof(SCLHdr));
    h.noSecs = hoHdr.noSecs;
    exitCode = writeFile(h, file, folderNum);
  }
  if(type == FMT_SCL)
  {
    SetFilePointer(file, 8, NULL, FILE_BEGIN);
    BYTE no_files;
    
    ReadFile(file, &no_files, 1, &noBytesRead, 0);

    SCLHdr* hdrs = new SCLHdr[no_files]{};
    ReadFile(file, hdrs, no_files*sizeof(SCLHdr), &noBytesRead, NULL);

    for(int fNum = 0; fNum < no_files; ++fNum)
    {
      ExitCode code = writeFile(hdrs[fNum], file, folderNum);

      if(code != OK) exitCode = code;
      if(code == CANCEL) break;
    }
    delete[] hdrs;
  }
  if(type == FMT_PLAIN)
  {
    DWORD fileSize = GetFileSize(file, NULL);;
    int sizeInSec  = fileSize/sectorSize  + (fileSize%sectorSize ? 1 : 0);
    int no_files   = sizeInSec/255 + !!(sizeInSec%255);
    if(!no_files) no_files = 1;
    
    char name[11] = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', 'C', 0, 0};
    if(no_files > 1) std::fill_n(name, 8, '_');
    
    make8x3name(pointToName(fileName), name);
    for(int fNum = 0; fNum < no_files; ++fNum)
    {
      SCLHdr hdr;
      
      if(no_files > 1)
      {
        char num[5];
        wsprintf(num, "_%03d", fNum);
        std::copy_n(num, 4, name + 4);
      }

      std::copy_n(name, std::size(hdr.name), hdr.name);
      memcpy(&hdr.type, name+8, 3);
      if(sizeInSec > 255)
      {
        hdr.noSecs = 255;
        hdr.size   = 255 * sectorSize;
        
        sizeInSec -= 255;
        fileSize  -= 255 * sectorSize;
      }
      else
      {
        //  
        hdr.noSecs = sizeInSec;
        hdr.size   = fileSize;
      }
      ExitCode code = writeFile(hdr, file, folderNum);

      if(code != OK) exitCode = code;
      if(code == CANCEL) break;
    }
  }
  
  CloseHandle(file);
  if(exitCode == OK && move) DeleteFileW(fileName);
  return exitCode;
}

bool Manager::makeFolder(const wchar_t* folderName, BYTE folderNum)
{
  if(!op.useDS) return false;
  if(!dsOk)
  {
    const wchar_t * const msgItems[]
    {
        GetMsg(MWarning), GetMsg(MDSNotInstalled), GetMsg(MInstallDS),
        GetMsg(MOk), GetMsg(MCancel)
    };
    if(messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 2) != 0) return false;
    dsOk = true;
  }
  if(noFolders == 127)
  {
    errorMessage(GetMsg(MTooManyFolders));
    return false;
  }

  std::fill(std::begin(folders[noFolders]), std::end(folders[noFolders]), ' ');

  make8x3name(folderName, folders[noFolders]);
  folderMap[noFolders]    = folderNum;
  //       
  pcFolders[noFolders][0] = 0; 
  if(folders[noFolders][0] == 0x01) ++noDelFolders;
  ++noFolders;
  return true;
}

ExitCode Manager::putFolder(const wchar_t* folderName, BYTE folderNum, bool move)
{
  int fNum = 0;
  const wchar_t* name = pointToName(folderName);
  for(; fNum < noFolders; ++fNum)
  {
    if(folderMap[fNum] != folderNum) continue;
    if(!lstrcmpW(pcFolders[fNum], name)) break;
  }
  if(fNum == noFolders)
  {
    //      
    if(!makeFolder(name, folderNum)) return CANCEL;
  }

  wchar_t mask[300];
  lstrcpyW(mask, folderName);
  addEndSlash(mask);
  lstrcatW(mask, L"*.*");

  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(mask, &data);
  if(h == INVALID_HANDLE_VALUE) return OK;

  ExitCode exitCode = OK;
  do
  {
    if(!lstrcmpW(data.cFileName, L".")) continue;
    if(!lstrcmpW(data.cFileName, L"..")) continue;

    ExitCode code;
    
    wchar_t fullName[300];
    lstrcpyW(fullName, folderName);
    addEndSlash(fullName);
    lstrcatW(fullName, data.cFileName);
    
    if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      code = putFolder(fullName, fNum+1, move);
    else
      code = putFile(fullName, fNum+1, move);

    if(code != OK) exitCode = code;
    if(code == CANCEL) break;
    
  }while(FindNextFileW(h, &data));
  FindClose(h);

  if(move && exitCode == OK)
    if(!RemoveDirectoryW(folderName)) return SKIP;
  return exitCode;
}

int Manager::putFiles(PluginPanelItem *panelItem, int noItems, int move, int opMode)
{
  if(noItems == 0) return 1;
  if(fmt->isProtected(img))
  {
    const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MCanNotCopy), GetMsg(MWriteProtected), GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }
  
  if(!openHostFile()) return 0;
  
  int returnCode   = 1;
  userAction       = ASK_USER;
  keepSilence      = opMode & OPM_SILENT;
  
  HANDLE file;
  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);

  size_t Size = startupInfo.PanelControl(PANEL_ACTIVE, FCTL_GETPANELDIRECTORY, 0, 0);
  std::unique_ptr<FarPanelDirectory, decltype(::free) *> dirInfo((FarPanelDirectory *)::malloc(Size), ::free);

  dirInfo->StructSize = sizeof(FarPanelDirectory);
  startupInfo.PanelControl(PANEL_ACTIVE, FCTL_GETPANELDIRECTORY, Size, dirInfo.get());

  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    ExitCode exitCode;

    std::wstring FullPath(dirInfo->Name);
    addEndSlash(FullPath);
    FullPath.append(panelItem[iNum].FileName);

    if(panelItem[iNum].FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      exitCode = putFolder(FullPath.c_str(), curFolderNum, move);
    else
      exitCode = putFile(FullPath.c_str(), curFolderNum, move);
    
    if(exitCode == SKIP)
    {
      returnCode = 0;
      continue;
    }
    if(exitCode == CANCEL)
    {
      returnCode = -1;
      break;
    }
    //  /   
    panelItem[iNum].Flags ^= PPIF_SELECTED;
  }
  startupInfo.RestoreScreen(screen);
  diskInfo.noFiles     = noFiles;
  diskInfo.noDelFiles  = noDelFiles;

  writeInfo();
  closeHostFile();
  op.reread = true;
  startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
  op.reread = false;
  return returnCode;
}
